require 'global_phone'

GlobalPhone.db_path = ENV.fetch('NUMBER_DB_FILE', 'phone_numbers.json')

module Ferryman
  class ConfigError < StandardError
  end

  class DeliveryError < StandardError
  end

  class RoutingRule
    attr_reader :strategy
    attr_reader :providers

    MATCH_ALL = lambda {|_| true}
    STRATEGIES = {
      'first' => lambda {|providers| providers},
      'random' => lambda {|providers| providers.shuffle}
    }

    def initialize
      @providers = []
      @match_rules = []
      @strategy = 'first'
    end

    def accept?(phone_number)
      matched = true

      @match_rules.each do |rule|
        matched &&= rule.call(phone_number)
        break unless matched
      end

      matched
    end

    def set_match(match)
      raise ConfigError.new('match cannot be nil!') if match.nil?

      if match == '*'
        @match_rules << MATCH_ALL
      elsif match.respond_to? 'has_key?'
        if match.has_key?(:regions)
          regions = Array.new(match[:regions])
          @match_rules << lambda do |number|
            parsed_number = GlobalPhone.parse(number)
            regions.include? parsed_number.territory.name
          end
        end
      else
        raise ConfigError.new("Unexpected value for 'match'")
      end

      true
    end

    def providers
      STRATEGIES[@strategy].call(@providers)
    end

    def add_provider(provider)
      raise ConfigError.new('provider cannot be nil!') if provider.nil?
      @providers << provider
    end

    def strategy=(strategy)
      raise ConfigError.new('unknown strategy') unless STRATEGIES.has_key?(strategy)
      @strategy = strategy
    end

  end

  class Config
    attr_reader :rules, :providers

    def initialize
      @providers = {}
      @rules = []
      @arity = nil
    end

    def add_provider(name, provider)
      raise ConfigError.new("Provider #{name} already exists") if @providers.has_key?(name)
      raise ConfigError.new("Provider must accept at least a phone number") if provider.arity < 1
      raise ConfigError.new("Provider arguments does not match with existing") unless @arity.nil? || provider.arity == @arity

      @arity = provider.arity
      @providers[name] = provider

      true
    end

    def set_routing_table(table)
      table.each do |row|
        raise ConfigError.new('providers cannot be nil!') if row[:providers].nil?

        rule = RoutingRule.new
        rule.set_match row[:match]
        rule.strategy = row.fetch(:strategy, 'first')
        row[:providers].each {|provider| rule.add_provider(provider)}
        @rules << rule
      end

      true
    end
  end

  class Dispatcher
    attr_reader :config

    def initialize
      @config = Config.new
    end

    def deliver(phone_number, *args)
      @config.rules.each do |rule|
        next unless rule.accept?(phone_number)

        rule.providers.each do |provider_name|
          provider = @config.providers[provider_name]
          params = [phone_number] + args
          begin
            return provider.call(*params)
          rescue DeliveryError
            next
          end
        end
      end

      nil
    end
  end

  def Ferryman.create
    dispatcher = Dispatcher.new
    yield dispatcher.config
    dispatcher
  end
end
